﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Xml;
using System.Xml.Linq;

namespace FieldEdit
{

	// Uses 10.0 feature service functionality (not replica-aware)
	// ASSUMPTIONS:
	// 1) Records created or updated in the field are never deleted on the server
	// 2) Geometry types to be synchronized are 2D Points, Polylines, or Polygons
	// 3) Classes to be synchronized must have both IDSTAMP (string) and TIMESTAMP (datetime) fields
	//    [GLOBALID cannot be used because a feature service cannot add a feature with a given global id]
	// 4) Relationships between classes to be synchronized are through the IDSTAMP field

	public class SyncConstants
	{
		public const string IDField = "IDSTAMP";
		public const string DateField = "TIMESTAMP";
		public const string SyncName = "Synchronization";
		public static string[] SyncLayers = {
			"RedPoint",
			"RedLine",
			"RedPoly",
			"Measurements",
			"Casing",
			"CP Insulated Facility",
			"CP Test Connection",
			"CP_Casing_Note",
			"CP_Casing_Reading",
			"CP_Circuit_Note",
			"CP_Insulated_Facility_Note",
			"CP_Insulated_Facility_Reading",
			"CP_Test_Reading"};
		public const string FieldEditLog = "c:/apps/FieldEdit/Data/FieldEdits.xml";
		public const string SyncServer = "http://tuswtesri01/ArcGIS/rest/services/";
		public static string[] SyncServices = { "Field_ESRD", "Field_TELD" };
	}

	public enum SyncLayerType
	{
		// This is expected to match SyncConstants.SyncLayers
		RedPoint = 0,
		RedLine = 1,
		RedPoly = 2,
		Measurements = 3,
		Casing = 4,
		CPInsulatedFacility = 5,
		CPTestConnection = 6,
		CPCasingNote = 7,
		CPCasingReading = 8,
		CPCircuitNote = 9,
		CPInsulatedFacilityNote = 10,
		CPInsulatedFacilityReading = 11,
		CPTestReading = 12
	}

	public enum ActivityType
	{
		// This is expected to match SyncConstants.SyncServices
		Gas = 0,
		Electric = 1
	}

	// A feature is an XElement object consisting of 
	// the XML equivalent of the JSON feature specification

	public struct UpdateSet
	{
		public Dictionary<string, string> FieldInfo;
		public List<XElement> AddFeatureList;
		public List<XElement> UpdateFeatureList;
	}

	public struct ActionSet
	{
		public List<string> AddIDList;
		public List<string> UpdateIDList;
	}

	public interface ISyncStatus
	{
		void Status(string sMsg);
		void Message(string sMsg);
		void Error(string sErr);
	}

	public interface ISynchronize
	{
		void SetStatus(ISyncStatus pStatus);
		// SetStatus provides an optional way to send status updates to a GUI
		void SetTimestamp(DateTime dtTimestamp);
		// Sets timestamp used to send edits to the remote server
		int GetUpdates(DateTime SyncDate, ISynchronize pTarget);
		// Used to get updates from server feature service
		// If SyncDate == DateTime.MinVal, all features are returned
		int GetUpdates(Dictionary<SyncLayerType, ActionSet> Actions, ISynchronize pTarget);
		// Used to get updates from local feature service
		int ProcessFeatures(Dictionary<SyncLayerType, UpdateSet> Features);
	}

	public class clsFieldHelper
	{

		private ActivityType _ActivityType;
		private List<SyncLayerType> _SyncTypes;
		private ISynchronize _pLocalSync;
		private ISynchronize _pServerSync;
		private ISyncStatus _pStatus;

		#region Public Methods
		public void Initialize(ActivityType actType, string sLocalServiceUrl, List<SyncLayerType> SyncTypes, ISyncStatus pStatus)
		{
			_ActivityType = actType;
			_SyncTypes = SyncTypes;
			_pStatus = pStatus;
			_pLocalSync = new clsSyncFeatureService(true, sLocalServiceUrl, _SyncTypes);
			_pLocalSync.SetStatus(_pStatus);
			string sServiceURL = SyncConstants.SyncServer;
			sServiceURL += SyncConstants.SyncServices[(int)actType];
			_pServerSync = new clsSyncFeatureService(false, sServiceURL, _SyncTypes);
			_pServerSync.SetStatus(_pStatus);
		}

		public static string PlacementCode(string sPC)
		{
			string sUser = Environment.GetEnvironmentVariable("USERNAME");
			string sPCode = sUser + ":" + DateTime.Now.ToString("MMddyy");
			if (sPC == "")
				return sPCode;
			if (sPC.IndexOf(sPCode) > -1)
				return sPCode;
			sPCode = sPC + "|" + sPCode;
			if (sPCode.Length > 30)
				sPCode = sPCode.Substring(sPCode.Length - 30);
			return sPCode;
		}

		public bool ProcessLocalFeature(SyncLayerType slType, XElement xFeat, bool bAdd)
		{
			Dictionary<SyncLayerType, UpdateSet> d = new Dictionary<SyncLayerType, UpdateSet>();
			UpdateSet updateSet = new UpdateSet();
			if (bAdd)
			{
				updateSet.AddFeatureList = new List<XElement>();
				updateSet.AddFeatureList.Add(xFeat);
			}
			else
			{
				updateSet.UpdateFeatureList = new List<XElement>();
				updateSet.UpdateFeatureList.Add(xFeat);
			}
			d[slType] = updateSet;
			bool bSuccess = (_pLocalSync.ProcessFeatures(d) > 0);
			string sLayer = SyncConstants.SyncLayers[(int)slType];
			if (bSuccess)
			{
				this.UpdateLog(slType, xFeat, bAdd);
				if (bAdd)
					this.Message(sLayer + " added.");
				else
					this.Message(sLayer + " updated.");
			}
			else
				this.Error("Could not add or update " + sLayer + ".");
			return bSuccess;
		}

		public void Synchronize()
		{

			// Receive server updates

			DateTime dtSync = this.GetSyncDate();
			DateTime dtNewSync = DateTime.Now;
			int iTotalReceived = _pServerSync.GetUpdates(dtSync, _pLocalSync);
			this.UpdateSyncDate(dtNewSync);

			// Get and send local updates

			Dictionary<SyncLayerType, ActionSet> Actions = this.GetUpdateActions();
			int iNumUpdates = 0;
			foreach (SyncLayerType slType in Actions.Keys)
			{
				ActionSet actionSet = Actions[slType];
				if (actionSet.AddIDList != null)
					iNumUpdates += actionSet.AddIDList.Count;
				if (actionSet.UpdateIDList != null)
					iNumUpdates += actionSet.UpdateIDList.Count;
			}
			int iTotalSent = 0;
			if (iNumUpdates > 0)
			{
				DateTime dtTimestamp = dtNewSync - new TimeSpan(0, 1, 0);
				_pServerSync.SetTimestamp(dtTimestamp);
				iTotalSent = _pLocalSync.GetUpdates(Actions, _pServerSync);
				if (iTotalSent == iNumUpdates)
					this.RemoveActions();
			}

			// Done

			string sMsg = "Features sent: " + iTotalSent.ToString();
			sMsg += "\nFeatures received: " + iTotalReceived.ToString();
			this.Status("OK");
			this.Message(sMsg);

		}
		#endregion

		#region Private Methods
		private void Status(string sMsg)
		{
			if (_pStatus == null)
				return;
			_pStatus.Status(sMsg);
		}

		private void Message(string sMsg)
		{
			if (_pStatus == null)
				return;
			_pStatus.Message(sMsg);
		}

		private void Error(string sErr)
		{
			if (_pStatus == null)
				return;
			_pStatus.Error(sErr);
		}

		private DateTime GetSyncDate()
		{

			// Open log and get sync node

			XDocument xEditLog = this.OpenEditLog();
			XElement xSync = this.FindSyncDateElement(xEditLog);

			// Get sync date

			DateTime dtSync = DateTime.MinValue;
			string sDate = this.GetAttribute(xSync, "date");
			if (sDate != null)
				dtSync = Convert.ToDateTime(sDate);
			return dtSync;

		}

		private void UpdateSyncDate(DateTime dtSync)
		{

			// Open log and get sync node

			XDocument xEditLog = this.OpenEditLog();
			XElement xSync = this.FindSyncDateElement(xEditLog);

			// Update sync date

			xSync.SetAttributeValue("date", dtSync.ToString());
			xEditLog.Save(SyncConstants.FieldEditLog);
		
		}

		private Dictionary<SyncLayerType, ActionSet> GetUpdateActions()
		{

			Dictionary<SyncLayerType, ActionSet> d = new Dictionary<SyncLayerType, ActionSet>();
			XDocument xEditLog = this.OpenEditLog();
			XElement xRoot = xEditLog.Root;
			foreach (XElement xParent in xRoot.Elements())
			{

				// Get Layer type
				
				string sElementName = xParent.Name.LocalName;
				if (sElementName != "layer")
					continue;
				string sName = this.GetAttribute(xParent, "name");
				SyncLayerType slType = SyncLayerType.RedPoint;
				int i = 0;
				foreach (string sLayerName in SyncConstants.SyncLayers)
				{
					if (sLayerName == sName)
					{
						slType = (SyncLayerType)i;
						break;
					}
					i++;
				}

				// Build actions

				ActionSet actionSet = new ActionSet();
				actionSet.AddIDList = new List<string>();
				actionSet.UpdateIDList = new List<string>();
				i = 0;
				foreach (XElement xChild in xParent.Elements())
				{
					string sID = this.GetAttribute(xChild, "id");
					string sAction = this.GetAttribute(xChild, "action");
					if (sAction == "add")
						actionSet.AddIDList.Add(sID);
					else
						actionSet.UpdateIDList.Add(sID);
					i++;
				}
				if (i == 0)
					continue;
				d[slType] = actionSet;

			}
			return d;

		}

		private void RemoveActions()
		{

			XDocument xEditLog = this.OpenEditLog();
			XElement xRoot = xEditLog.Root;
			List<XElement> ChildList = new List<XElement>();
			foreach (XElement xParent in xRoot.Elements())
			{

				// Get Layer

				string sElementName = xParent.Name.LocalName;
				if (sElementName != "layer")
					continue;

				// Get entries to delete

				foreach (XElement xChild in xParent.Elements())
					ChildList.Add(xChild);

			}

			// Remove actions from log

			foreach (XElement xChild in ChildList)
				xChild.Remove();

			// Save updated log

			xEditLog.Save(SyncConstants.FieldEditLog);

		}

		private XDocument OpenEditLog()
		{
			XDocument xEditLog = new XDocument();
			string sFieldEditLog = SyncConstants.FieldEditLog;
			if (File.Exists(sFieldEditLog))
			{
				XmlReader xr = XmlReader.Create(sFieldEditLog);
				xEditLog = XDocument.Load(xr);
				xr.Close();
			}
			else
			{
				XElement xElement = new XElement("FieldEdits");
				xEditLog.Add(xElement);
			}
			return xEditLog;
		}

		private XElement FindSyncDateElement(XDocument xEditLog)
		{

			// [The synchronization date node is named after SyncConstants.SyncName]

			XElement xRoot = xEditLog.Root;
			foreach (XElement xParent in xRoot.Elements())
			{
				if (xParent.Name.LocalName == SyncConstants.SyncName)
					return xParent;
			}
			XElement xNewElement = new XElement(SyncConstants.SyncName);
			xRoot.Add(xNewElement);
			return xNewElement;
		}

		private XElement FindLayerElement(XDocument xEditLog, string sLayerName)
		{

			// [Layer nodes are named "layer", have the attribute "name", and contain feature nodes]

			XElement xRoot = xEditLog.Root;
			foreach (XElement xParent in xRoot.Elements())
			{
				if (xParent.Name.LocalName != "layer")
					continue; // This should never happen
				if (this.GetAttribute(xParent, "name") == sLayerName)
					return xParent;
			}
			XElement xNewElement = new XElement("layer");
			xNewElement.SetAttributeValue("name", sLayerName);
			xRoot.Add(xNewElement);
			return xNewElement;
		}

		private string GetAttribute(XElement xElement, string sName)
		{
			foreach (XAttribute xAtt in xElement.Attributes())
			{
				if (xAtt.Name.LocalName == sName)
					return xAtt.Value;
			}
			return null;
		}

		private void UpdateLog(SyncLayerType slType, XElement xFeat, bool bNew)
		{

			// Get ID

			string sFeatID = "";
			foreach (XElement xChild in xFeat.Elements())
			{
				string sName = xChild.Name.LocalName;
				if (sName != "attributes")
					continue;
				foreach (XElement xAttr in xChild.Elements())
				{
					string sAttr = xAttr.Name.LocalName.ToUpper();
					if (sAttr != SyncConstants.IDField)
						continue;
					sFeatID = xAttr.Value;
				}
				break;
			}
			if (sFeatID == "")
				return;
			string sAction;
			if (bNew)
				sAction = "add";
			else
				sAction = "update";

			// Update log

			string sLayerName = SyncConstants.SyncLayers[(int)slType];
			XDocument xEditLog = OpenEditLog();

			// Get parent node for feature type

			XElement xParent = this.FindLayerElement(xEditLog, sLayerName);

			// [Feature nodes are named "feature" and have attributes "action" and "id"]
			// [action can have the value "add" or "update"]

			// If a feature is updated, check if an entry already exists

			if (!bNew)
			{

				XElement xPrevious = null;
				foreach (XElement xElement in xParent.Elements())
				{
					string sID = this.GetAttribute(xElement, "id");
					if (sID == sFeatID)
					{
						xPrevious = xElement;
						break;
					}
				}
				if (xPrevious != null)
					return;
			}

			// Add feature

			XElement xFeatElement = new XElement("feature");
			xFeatElement.SetAttributeValue("action", sAction);
			xFeatElement.SetAttributeValue("id", sFeatID);
			xParent.Add(xFeatElement);

			// Save updated log

			xEditLog.Save(SyncConstants.FieldEditLog);

		}
		#endregion

	}
}
