﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Threading;
using System.Xml.Linq;
using System.Windows;
using System.Windows.Input;
using System.Windows.Threading;

using ESRI.ArcGIS.Client;
using ESRI.ArcGIS.Client.Geometry;
using ESRI.ArcGIS.Client.Local;

using FieldEdit;

namespace FieldTest
{

	public partial class MainWindow : Window, ISyncStatus
	{

		private const string sPkgDir = "C:/apps/FieldEdit/packages/";
		private const string sOpsMapPkg = sPkgDir + "AZ Gas.mpk";
		private const string sFieldMapPkg = sPkgDir + "FieldEdit.mpk";

		private LocalMapService _localMapService;
		private ArcGISLocalDynamicMapServiceLayer _lyrOpsMap;
		private LocalFeatureService _localFeatureService;
		private ArcGISLocalDynamicMapServiceLayer _lyrFieldMap;
		private clsFieldHelper _FieldHelper;
		private Draw _EditDraw;

		public MainWindow()
		{
			// License setting and ArcGIS Runtime initialization is done in Application.xaml.cs.

			InitializeComponent();
			this.InitializeApp();
		}

		#region Initialization
		private void InitializeApp()
		{
			this.Status("Initializing map...");
			btnSynchronize.Click += new RoutedEventHandler(btnSynchronize_Click);
			btnAddPoint.Click += new RoutedEventHandler(btnAddPoint_Click);
			btnAddLine.Click += new RoutedEventHandler(btnAddLine_Click);
			btnAddPoly.Click += new RoutedEventHandler(btnAddPoly_Click);
			_localMapService = new LocalMapService(sOpsMapPkg);
			_localMapService.EnableDynamicLayers = true;
			_localMapService.StartCompleted += new EventHandler<System.ComponentModel.AsyncCompletedEventArgs>(_localMapService_StartCompleted);
			_localMapService.StartAsync();
			_localFeatureService = new LocalFeatureService(sFieldMapPkg);
			_localFeatureService.StartCompleted += new System.EventHandler<System.ComponentModel.AsyncCompletedEventArgs>(_localFeatureService_StartCompleted);
			_localFeatureService.StartAsync();

		}

		private void _localMapService_StartCompleted(object sender, System.ComponentModel.AsyncCompletedEventArgs e)
		{
			_lyrOpsMap = new ArcGISLocalDynamicMapServiceLayer(_localMapService);
			_lyrOpsMap.ID = "AZ Gas";
			_lyrOpsMap.Initialized += new System.EventHandler<System.EventArgs>(_lyrOpsMap_Initialized);
			_lyrOpsMap.InitializationFailed += new EventHandler<EventArgs>(_lyrOpsMap_InitializationFailed);
			_lyrOpsMap.Initialize();
		}

		private void _localFeatureService_StartCompleted(object sender, System.ComponentModel.AsyncCompletedEventArgs e)
		{
			_lyrFieldMap = new ArcGISLocalDynamicMapServiceLayer(_localFeatureService);
			_lyrFieldMap.ID = "Field Data";
			_lyrFieldMap.Initialized += new EventHandler<EventArgs>(_lyrFieldMap_Initialized);
			_lyrFieldMap.InitializationFailed += new EventHandler<EventArgs>(_lyrFieldMap_InitializationFailed);
			_lyrFieldMap.Initialize();
		}

		void _lyrOpsMap_Initialized(object sender, EventArgs e)
		{
			if (_lyrFieldMap == null)
				return;
			if (!_lyrFieldMap.IsInitialized)
				return;
			this.FinishInit();
		}

		private void _lyrOpsMap_InitializationFailed(object sender, EventArgs e)
		{
			MessageBox.Show("Ops layer failed to initialize.");
		}

		void _lyrFieldMap_InitializationFailed(object sender, EventArgs e)
		{
			MessageBox.Show("Field layer failed to initialize.");
		}

		void _lyrFieldMap_Initialized(object sender, EventArgs e)
		{
			if (_lyrOpsMap == null)
				return;
			if (!_lyrOpsMap.IsInitialized)
				return;
			this.FinishInit();
		}

		void FinishInit()
		{
			_Map.Layers.Add(_lyrOpsMap);
			_Map.Layers.Add(_lyrFieldMap);
			_Legend.Map = _Map;
			_Legend.LayerTemplate = (DataTemplate)this.Resources["My_LayerTemplate"];
			_Legend.LayerIDs = new string[] { "Field Data", "AZ Gas" };
			List<SyncLayerType> SyncTypes = new List<SyncLayerType>();
			SyncTypes.Add(SyncLayerType.RedPoint);
			SyncTypes.Add(SyncLayerType.RedLine);
			SyncTypes.Add(SyncLayerType.RedPoly);
			_FieldHelper = new clsFieldHelper();
			_FieldHelper.Initialize(ActivityType.Gas, _localFeatureService.UrlFeatureService, SyncTypes, this);
			_EditDraw = new Draw(_Map);
			_EditDraw.DrawMode = DrawMode.Point;
			_EditDraw.DrawComplete += new System.EventHandler<DrawEventArgs>(_EditDraw_DrawComplete);
			_EditDraw.IsEnabled = false;
			this.Status("OK");
		}
		#endregion

		public void Status(string sMsg)
		{
			tbStatus.Text = sMsg;
			tbStatus.InvalidateVisual();
			// This is a WPF equivalent of Application.DoEvents()
			Application.Current.Dispatcher.Invoke(DispatcherPriority.Background, new ThreadStart(delegate { }));
		}

		public void Message(string sMsg)
		{
			MessageBox.Show(sMsg);
		}
		
		public void Error(string sErr)
		{
			tbStatus.Text = "ERROR";
			MessageBox.Show(sErr, "ERROR", MessageBoxButton.OK, MessageBoxImage.Error);
		}

		void btnSynchronize_Click(object sender, RoutedEventArgs e)
		{
			_FieldHelper.Synchronize();
			_lyrFieldMap.Refresh();
		}

		void btnAddPoint_Click(object sender, RoutedEventArgs e)
		{
			_Map.Cursor = Cursors.Cross;
			_EditDraw.DrawMode = DrawMode.Point;
			_EditDraw.IsEnabled = true;
		}

		void btnAddLine_Click(object sender, RoutedEventArgs e)
		{
			_Map.Cursor = Cursors.Cross;
			_EditDraw.DrawMode = DrawMode.Polyline;
			_EditDraw.IsEnabled = true;
		}

		void btnAddPoly_Click(object sender, RoutedEventArgs e)
		{
			_Map.Cursor = Cursors.Cross;
			_EditDraw.DrawMode = DrawMode.Polygon;
			_EditDraw.IsEnabled = true;
		}

		void _EditDraw_DrawComplete(object sender, DrawEventArgs e)
		{
			_EditDraw.IsEnabled = false;
			_Map.Cursor = Cursors.Arrow;
			frmMarkup frm = new frmMarkup();
			Dictionary<string, object> attrs = frm.Open();
			if (attrs == null)
				return;
			Geometry g = e.Geometry;
			SyncLayerType slType = SyncLayerType.RedPoint;
			if (g is Polyline)
				slType = SyncLayerType.RedLine;
			else if (g is Polygon)
				slType = SyncLayerType.RedPoly;
			XElement xFeat = this.CreateXElement(g, attrs);
			_FieldHelper.ProcessLocalFeature(slType, xFeat, true);
			_lyrFieldMap.Refresh();

		}

		private XElement CreateXElement(Geometry g, Dictionary<string, object> attrs)
		{
			XElement xFeat = new XElement("feature");
			xFeat.SetAttributeValue("type", "object");
			if (g != null)
				xFeat.Add(GetXGeometry(g));
			xFeat.Add(GetXAttributes(attrs));
			return xFeat;
		}

		private XElement GetXGeometry(Geometry g)
		{
			XElement xGeom = new XElement("geometry");
			xGeom.SetAttributeValue("type", "object");
			if (g is MapPoint)
			{
				MapPoint pt = (MapPoint)g;
				XElement xX = new XElement("x");
				xX.SetAttributeValue("type", "number");
				xX.Value = pt.X.ToString();
				xGeom.Add(xX);
				XElement xY = new XElement("y");
				xY.SetAttributeValue("type", "number");
				xY.Value = pt.Y.ToString();
				xGeom.Add(xY);
			}
			else if (g is Envelope)
			{
				Envelope env = (Envelope)g;
				PointCollection PC = new PointCollection();
				PC.Add(new MapPoint(env.XMin, env.YMin));
				PC.Add(new MapPoint(env.XMin, env.YMax));
				PC.Add(new MapPoint(env.XMax, env.YMax));
				PC.Add(new MapPoint(env.XMax, env.YMin));
				PC.Add(new MapPoint(env.XMin, env.YMin));
				XElement xParts = new XElement("rings");
				xParts.SetAttributeValue("type", "array");
				XElement xPart = new XElement("item");
				xPart.SetAttributeValue("type", "array");
				foreach (MapPoint pt in PC)
				{
					XElement xPt = new XElement("item");
					xPt.SetAttributeValue("type", "array");
					XElement xX = new XElement("item");
					xX.SetAttributeValue("type", "number");
					xX.Value = pt.X.ToString();
					xPt.Add(xX);
					XElement xY = new XElement("item");
					xY.SetAttributeValue("type", "number");
					xY.Value = pt.Y.ToString();
					xPt.Add(xY);
					xPart.Add(xPt);
				}
				xParts.Add(xPart);
				xGeom.Add(xParts);
			}
			else
			{
				string sName = "";
				ObservableCollection<PointCollection> PCList = null;
				if (g is Polyline)
				{
					Polyline p = (Polyline)g;
					sName = "paths";
					PCList = p.Paths;
				}
				else if (g is Polygon)
				{
					Polygon p = (Polygon)g;
					sName = "rings";
					PCList = p.Rings;
				}
				else
					return xGeom;
				XElement xParts = new XElement(sName);
				xParts.SetAttributeValue("type", "array");
				foreach (PointCollection PC in PCList)
				{
					XElement xPart = new XElement("item");
					xPart.SetAttributeValue("type", "array");
					foreach (MapPoint pt in PC)
					{
						XElement xPt = new XElement("item");
						xPt.SetAttributeValue("type", "array");
						XElement xX = new XElement("item");
						xX.SetAttributeValue("type", "number");
						xX.Value = pt.X.ToString();
						xPt.Add(xX);
						XElement xY = new XElement("item");
						xY.SetAttributeValue("type", "number");
						xY.Value = pt.Y.ToString();
						xPt.Add(xY);
						xPart.Add(xPt);
					}
					xParts.Add(xPart);
				}
				xGeom.Add(xParts);
			}
			return xGeom;
		}

		private XElement GetXAttributes(Dictionary<string, object> attrs)
		{
			XElement xAttr = new XElement("attributes");
			xAttr.SetAttributeValue("type", "object");
			foreach (string sKey in attrs.Keys)
			{
				object oVal = attrs[sKey];
				string sVal = oVal.ToString();
				string sType = "string";
				if ((oVal is int) || (oVal is double))
					sType = "number";
				else if (oVal is DateTime)
				{
					// Convert to JSON value
					DateTime dt = (DateTime)oVal;
					DateTime epoch = new DateTime(1970, 1, 1, 0, 0, 0, 0);
					TimeSpan span = dt - epoch;
					double dm = span.TotalMilliseconds;
					sVal = dm.ToString("#");
					sType = "number";
				}
				XElement xElement = new XElement(sKey);
				xElement.SetAttributeValue("type", sType);
				xElement.Value = sVal;
				xAttr.Add(xElement);
			}
			return xAttr;
		}

	}
}
